/*
 *  Digital Audio (PCM) abstraction layer
 *  Copyright (c) by Jaroslav Kysela <perex@jcu.cz>
 */

#define ULTRA_MAIN_OBJECT_FILE
#include "driver.h"
#include "pcm.h"

#define ULTRA_PCM_DEVICES	2

static ultra_pcm_t *ultra_pcm_devices[ ULTRA_CARDS * ULTRA_PCM_DEVICES ];

#if 0
#define ULTRA_PCM_DEBUG_BUFFERS
#endif

int ultra_mixer_ioctl_card( ultra_card_t *card, struct file *file, unsigned int cmd, unsigned long arg );

static void ultra_pcm_fill_with_neutral( ultra_pcm_channel_t *pchn );
static void ultra_pcm_compute_blocks( ultra_pcm_t *pcm, ultra_pcm_channel_t *pchn );

/*
 *
 */

int ultra_pcm_dma_alloc( ultra_card_t *card, ultra_pcm_channel_t *pchn, int dmanum, char *ident, int lock )
{
  struct ULTRA_STRU_DMA *pdma;

  if ( ultra_dma_malloc( card, dmanum, ident, lock ) < 0 )
    return -EBUSY;
  pdma = &card -> dmas[ dmanum ];  
  pchn -> _dma = pdma;
  pchn -> dma = pdma -> dma;
  pchn -> size = pdma -> usize;
  pchn -> buffer = pdma -> buf;
  pchn -> frag_size = 0;
  pchn -> flags |= ULTRA_PCM_FLG_DMAOK;
  return 0;
}

int ultra_pcm_dma_free( ultra_card_t *card, ultra_pcm_channel_t *pchn, int dmanum, int lock )
{
  ultra_dma_free( card, dmanum, lock );
  pchn -> flags &= ~ULTRA_PCM_FLG_DMAOK;
  return 0;
}

static void ultra_pcm_clear_channel( ultra_pcm_channel_t *pchn )
{
  pchn -> voices = 1;
  pchn -> mode = 0;
  pchn -> format = 0;
  pchn -> rate = ULTRA_PCM_DEFAULT_RATE;
  pchn -> requested_block_size = 0;
  pchn -> requested_blocks = 0;
  pchn -> requested_subdivision = 0;
  pchn -> processed_bytes = 0;
  pchn -> interrupts = 0;
  pchn -> flags = 0;
  pchn -> used_size = 0;
  pchn -> neutral_byte = 0;
  pchn -> _dma = NULL;
  pchn -> hidden_buffer = NULL;
  pchn -> dma = 0;
  pchn -> size = 0;
  pchn -> buffer = NULL;
  pchn -> blocks = 0;
  pchn -> block_size = 0;
  pchn -> used = 0;
  pchn -> frag_size = 0;
  pchn -> head = 0;
  pchn -> tail = 0;
  ULTRA_PCM_LOCKZERO( pchn );
}

/*
 *
 */

static unsigned short ultra_pcm_file_flags( unsigned short f_flags )
{
  switch ( f_flags & O_ACCMODE ) {
    case O_WRONLY:	return ULTRA_PCM_LFLG_PLAY;
    case O_RDONLY:	return ULTRA_PCM_LFLG_RECORD;
    default:		return ULTRA_PCM_LFLG_BOTH;
  }
}

/*
 *  some shared routines for lowlevel driver
 */

void ultra_pcm_playback_dma( ultra_pcm_t *pcm,
			     unsigned char *buffer, unsigned int offset,
			     unsigned char *user, unsigned int count )
{
#if 0
  printk( "playback_dma: buffer = 0x%lx, offset = 0x%x, user = 0x%lx, count = 0x%x\n", (long)buffer, offset, (long)user, count );
#endif
  MEMCPY_FROMFS( &buffer[ offset ], user, count );
}

void ultra_pcm_playback_dma_neutral( ultra_pcm_t *pcm,
				     unsigned char *buffer, unsigned int offset,
				     unsigned int count,
				     unsigned char neutral_byte )
{
  MEMSET( &buffer[ offset ], neutral_byte, count );
}

void ultra_pcm_record_dma( ultra_pcm_t *pcm,
			   unsigned char *buffer, unsigned int offset,
			   unsigned char *user, unsigned int count )
{
  MEMCPY_TOFS( user, &buffer[ offset ], count );
}

/*
 *  interrupt callbacks from lowlevel driver
 */
 
void ultra_pcm_interrupt_playback( ultra_pcm_t *pcm )
{
  unsigned long flags;
  ultra_pcm_channel_t *pchn;
  int trigger = 0;
  
  pchn = &pcm -> playback;
  if ( !(pchn -> flags & ULTRA_PCM_FLG_TRIGGER) ) {
    PRINTD( "ultra_pcm_interrupt_playback: Oops, playback interrupt when playback isn't active!!!\n" );
    return;
  }
  if ( !(pchn -> flags & ULTRA_PCM_FLG_MMAP) ) {
    /* try clear first few bytes from next block */
    if ( pchn -> hw.flags & ULTRA_PCM_HW_AUTODMA ) {
      MEMSET( pchn -> buffer + pchn -> block_size * pchn -> tail, pchn -> neutral_byte, 16 );
    }
    if ( pchn -> used == 1 ) {
      if ( !(pchn -> flags & ULTRA_PCM_FLG_SYNC) ) {
#ifdef ULTRA_PCM_DEBUG_BUFFERS
        printk( "playback: underrun!!! - jiffies = %li\n", jiffies );
#endif
        pchn -> hw.discarded++;
      }
    }
    CLI( &flags );
    pchn -> interrupts++;
    if ( pchn -> used > 0 )
      pchn -> processed_bytes += pchn -> block_size;
    pchn -> tail++;
    pchn -> tail %= pchn -> blocks;
    pchn -> used--;
    if ( pchn -> used > 0 ) {
      ULTRA_PCM_LOCK( pchn, pchn -> tail );
    } else {
      ULTRA_PCM_LOCKZERO( pchn );
    }
    STI( &flags );
#if 0
    printk( "interrupt!!! tail = %i, used = %i\n", pchn -> tail, pchn -> used );
#endif
    if ( pchn -> used > 0 ) trigger = 1;
  } else {
    CLI( &flags );
    pchn -> interrupts++;
    pchn -> processed_bytes += pchn -> block_size;
    pchn -> tail++;
    pchn -> tail %= pchn -> blocks;
    STI( &flags );
    trigger = 1;
  }
  if ( trigger && !(pchn -> hw.flags & ULTRA_PCM_HW_AUTODMA) ) {
    pchn -> hw.prepare( pcm,
            		pchn -> buffer, pchn -> used_size,            			
            		pchn -> tail * pchn -> block_size,
                        pchn -> used_size );
    pchn -> hw.trigger( pcm, 1 );
  }
  if ( !trigger ) {
    if ( pchn -> hw.flags & ULTRA_PCM_HW_AUTODMA ) {
      pchn -> hw.trigger( pcm, 0 );
    }
    pchn -> flags &= ~ULTRA_PCM_FLG_TRIGGER;
  }
  CLI( &flags );
  if ( pchn -> flags & ULTRA_PCM_FLG_SLEEP ) {
    pchn -> flags &= ~ULTRA_PCM_FLG_SLEEP;
    WAKEUP( pcm, playback );
  }
  STI( &flags );
}

void ultra_pcm_interrupt_record( ultra_pcm_t *pcm )
{
  unsigned long flags;
  ultra_pcm_channel_t *pchn;
  int trigger;
  
  pchn = &pcm -> record;
  if ( !(pchn -> flags & ULTRA_PCM_FLG_TRIGGER) ) {
    PRINTD( "ultra_pcm_interrupt_record: Oops, record interrupt when record isn't active!!!\n" );
    return;
  }
  trigger = 0;
  if ( !(pchn -> flags & ULTRA_PCM_FLG_MMAP) ) {
    CLI( &flags );
    if ( pchn -> used < pchn -> blocks ) {
      pchn -> used++;
    } else {
      pchn -> hw.discarded++;
    }
    pchn -> interrupts++;
    pchn -> processed_bytes += pchn -> block_size;
    pchn -> head++;
    pchn -> head %= pchn -> blocks;
    ULTRA_PCM_LOCK( pchn, pchn -> head );
    STI( &flags );
  } else {
    CLI( &flags );
    pchn -> interrupts++;
    pchn -> processed_bytes += pchn -> block_size;
    pchn -> head++;
    pchn -> head %= pchn -> blocks;
    STI( &flags );
  }
  if ( !(pchn -> flags & ULTRA_PCM_FLG_SYNC) ) trigger = 1;
  if ( trigger && !(pchn -> hw.flags & ULTRA_PCM_HW_AUTODMA) ) {
    pchn -> hw.prepare( pcm,
            		pchn -> buffer, pchn -> used_size,            			
            		pchn -> head * pchn -> block_size,
                        pchn -> used_size );
    pchn -> hw.trigger( pcm, 1 );
  }
  if ( !trigger ) { 
    if ( pchn -> hw.flags & ULTRA_PCM_HW_AUTODMA ) {
      pchn -> hw.trigger( pcm, 0 );
    }
    pchn -> flags &= ~ULTRA_PCM_FLG_TRIGGER;
  }
  CLI( &flags );
  if ( pchn -> flags & ULTRA_PCM_FLG_SLEEP ) {
    pchn -> flags &= ~ULTRA_PCM_FLG_SLEEP;
    WAKEUP( pcm, record );
  }
  STI( &flags );
}

/*
 *  trigger standard buffered playback
 */

static void ultra_pcm_trigger_playback( ultra_pcm_t *pcm, ultra_pcm_channel_t *pchn )
{
  unsigned long flags;

  if ( !pchn -> used ) return;
  CLI( &flags );
  if ( !(pchn -> flags & ULTRA_PCM_FLG_TRIGGER) ) {
    STI( &flags );
    pchn -> flags |= ULTRA_PCM_FLG_TRIGGERA;
    pchn -> hw.prepare( pcm,
            		pchn -> buffer, pchn -> used_size,            			
            		pchn -> tail * pchn -> block_size,
                        pchn -> block_size );
    pchn -> hw.trigger( pcm, 1 );
    ULTRA_PCM_LOCK( pchn, pchn -> tail );
    CLI( &flags );
#if 0
    if ( !pchn -> processed_bytes ) {
      pchn -> interrupts++;
      pchn -> processed_bytes += pchn -> block_size;
    }
#endif
  }
  STI( &flags );
}

/*
 *  user to dma
 */

static int ultra_pcm_user_to_dma( ultra_pcm_t *pcm, const char *buf, int count )
{
  unsigned long flags;
  ultra_pcm_channel_t *pchn;
  int result, tmp, realtime;

  if ( count <= 0 || !buf ) return 0;
  if ( VERIFY_AREA( VERIFY_READ, buf, count ) ) return -EACCES;

  pchn = &pcm -> playback;
#if 0
  PRINTK( "pcm_user_to_dma: buf=0x%x, count=0x%x,%i (%s), used = %i, jiffies = %li\n", (int)buf, count, count, pchn -> flags & ULTRA_PCM_FLG_NONBLK ? "non blocked" : "blocked", pchn -> used, jiffies );
#endif
  count &= ~( (pchn -> voices * (pchn -> mode & ULTRA_PCM_MODE_16 ? 2 : 1)) - 1 );
  pchn -> flags &= ~ULTRA_PCM_FLG_ABORT;
  if ( pchn -> flags & ULTRA_PCM_FLG_MMAP ) return -EIO;	/* go to hell... */
  if ( pchn -> flags & ULTRA_PCM_FLG_BUFFERS )
    ultra_pcm_compute_blocks( pcm, pchn );
  if ( pchn -> flags & ULTRA_PCM_FLG_NEUTRAL )
    ultra_pcm_fill_with_neutral( pchn );

  result = 0;

  realtime = 0;
  if ( !(pchn -> flags & ULTRA_PCM_FLG_REALTM) &&
       !(pcm -> flags & ULTRA_PCM_LFLG_REALTIME) )
    realtime = 1;

  /*
   * OK. This is patch for .au files which begin with .snd header...
   * This is a little bit hard - it's application work to do conversions...
   */
  if ( (pchn -> mode & ULTRA_PCM_MODE_ULAW) && !pchn -> processed_bytes && !pchn -> used && count > 31 )
    {
      unsigned char buffer[ 8 ];
      
      MEMCPY_FROMFS( buffer, buf, 8 );
      if ( !memcmp( buffer, ".snd\0\0\0\x20", 8 ) )
        {
          buf += 32;
          count -= 32;
          result += 32;
        }
    }

  while ( count > 0 )
    {
      while ( pchn -> used >= pchn -> blocks )
        {
          if ( !(pchn -> flags & ULTRA_PCM_FLG_TRIGGER) )
            ultra_pcm_trigger_playback( pcm, pchn );
          if ( pchn -> flags & ULTRA_PCM_FLG_NONBLK ) return result;
          if ( !(pchn -> flags & ULTRA_PCM_FLG_ENABLE) ) return result;
          CLI( &flags );
          pchn -> flags |= ULTRA_PCM_FLG_SLEEP;
          SLEEP( pcm, playback, 10 * HZ );
          pchn -> flags &= ~ULTRA_PCM_FLG_SLEEP;
          STI( &flags );
          if ( TABORT( pcm, playback ) )
            {
              pchn -> flags |= ULTRA_PCM_FLG_ABORT;
              return result;
            }
          if ( pchn -> used >= pchn -> blocks && TIMEOUT( pcm, playback ) )
            {
              PRINTD( "pcm_user_to_dma: timeout, new block discarded\n" );
              return -EIO;
            }
        }
        
      /* ok.. interresant part.. we must find first free & not locked block.. */
        
      tmp = 0;
      CLI( &flags );
      while ( ULTRA_PCM_ISLOCK( pchn, pchn -> head ) )
        {
          pchn -> hw.discarded++;
#ifdef ULTRA_PCM_DEBUG_BUFFERS
          printk( "playback: discarded!!! %i - %i\n", pchn -> block_lock, pchn -> head );
#endif
          pchn -> used++;
          pchn -> head++;
          pchn -> head %= pchn -> blocks;
          if ( pchn -> used >= pchn -> blocks )
            {
              tmp++;
              break;
            }
        }
      STI( &flags );
      if ( tmp ) continue;		/* go to sleep loop */

      /* ok.. now we have right block - fill it */
      
      tmp = pchn -> block_size - pchn -> frag_size;
      if ( tmp > count ) tmp = count;	/* correction */

#if 0
      printk( "to dma: size = %i\n", tmp );
#endif
      if ( VERIFY_AREA( VERIFY_READ, buf, tmp ) )
        return -EACCES;
      pchn -> hw.dma( pcm,
                      pchn -> buffer,
                      ( pchn -> head * pchn -> block_size ) + pchn -> frag_size,
                      (char *)buf, tmp );

      count -= tmp;
      result += tmp;
      buf += tmp;

      CLI( &flags );
      pchn -> frag_size += tmp;
      if ( pchn -> frag_size >= pchn -> block_size )
        {
#if 0
          printk( "block filled = %i, used = %i, blocks = %i\n", pchn -> head, pchn -> used, pchn -> blocks );
#endif
          pchn -> used++;
          pchn -> head++;
          pchn -> head %= pchn -> blocks;
          pchn -> processed_bytes += pchn -> frag_size;
          pchn -> frag_size = 0;
        }
      STI( &flags );
    }
#if 0
  PRINTK( "pcm_user_to_dma: end\n" );
#endif
  if ( pchn -> used )
    ultra_pcm_trigger_playback( pcm, pchn );
  return result;
}

/*
 *  trigger standard buffered record
 */

static void ultra_pcm_trigger_record( ultra_pcm_t *pcm, ultra_pcm_channel_t *pchn )
{
  unsigned long flags;

  if ( pchn -> used >= pchn -> blocks ) return;
  CLI( &flags );
  if ( !(pchn -> flags & ULTRA_PCM_FLG_TRIGGER) ) {
    STI( &flags );
    pchn -> flags |= ULTRA_PCM_FLG_TRIGGERA;
    pchn -> hw.prepare( pcm,
            		pchn -> buffer, pchn -> used_size,            			
            		pchn -> head * pchn -> block_size,
                        pchn -> block_size );
    pchn -> hw.trigger( pcm, 1 );
    ULTRA_PCM_LOCK( pchn, pchn -> head );
    CLI( &flags );
#if 0
    if ( !pchn -> processed_bytes ) {
      pchn -> interrupts++;
      pchn -> processed_bytes += pchn -> block_size;
    }
#endif
  }
  STI( &flags );
}

/*
 *  dma to user
 */

static int ultra_pcm_dma_to_user( ultra_pcm_t *pcm, char *buf, int count )
{
  unsigned long flags;
  ultra_pcm_channel_t *pchn;
  int result, tmp;

  if ( count <= 0 ) return 0;
  if ( VERIFY_AREA( VERIFY_WRITE, buf, count ) ) return -EACCES;
  
  pchn = &pcm -> record;
#if 0
  PRINTK( "pcm_user_from_dma: buf=0x%x, count=0x%x (%s)\n", (int)buf, count, pchn -> flags & ULTRA_PCM_FLG_NONBLK ? "non blocked" : "blocked" );
#endif
  count &= ~( (pchn -> voices * (pchn -> mode & ULTRA_PCM_MODE_16 ? 2 : 1)) - 1 );
  pchn -> flags &= ~ULTRA_PCM_FLG_ABORT;
  if ( pchn -> flags & ULTRA_PCM_FLG_MMAP ) return -EIO;
  if ( pchn -> flags & ULTRA_PCM_FLG_BUFFERS )
    ultra_pcm_compute_blocks( pcm, pchn );
  if ( pchn -> flags & ULTRA_PCM_FLG_NEUTRAL )
    ultra_pcm_fill_with_neutral( pchn );

  if ( !(pchn -> flags & ULTRA_PCM_FLG_ENABLE) ) return 0;
  ultra_pcm_trigger_record( pcm, pchn );
  result = 0;
  
  while ( count > 0 )
    {
      while ( !pchn -> used )
        {
          if ( pchn -> flags & ULTRA_PCM_FLG_NONBLK ) return result;
          CLI( &flags );
          pchn -> flags |= ULTRA_PCM_FLG_SLEEP;
          SLEEP( pcm, record, 10 * HZ );
          pchn -> flags &= ~ULTRA_PCM_FLG_SLEEP;
          STI( &flags );
          if ( TABORT( pcm, record ) )
            {
              pchn -> flags |= ULTRA_PCM_FLG_ABORT;
              return -EINTR;
            }
          if ( !pchn -> used && TIMEOUT( pcm, record ) )
            {
              PRINTD( "ultra_pcm_dma_to_user: data timeout\n" );
              return -EIO;
            }
        }
      
      tmp = 0;
      CLI( &flags );
      while ( ULTRA_PCM_ISLOCK( pchn, pchn -> tail ) )
        {
          pchn -> hw.discarded++;
          pchn -> used--;
          pchn -> tail++;
          pchn -> tail %= pchn -> blocks;
          if ( !pchn -> used )
            {
              tmp++;
              break;
            }
        }
      STI( &flags );
      if ( tmp ) continue;		/* go to sleep loop */
        
      tmp = count <= pchn -> block_size ? count : pchn -> block_size;
      if ( VERIFY_AREA( VERIFY_WRITE, buf, tmp ) )
        return -EACCES;
      pchn -> hw.dma( pcm,
                      pchn -> buffer,
      		      ( pchn -> tail * pchn -> block_size ) + ( pchn -> block_size - pchn -> frag_size ),
                      buf, tmp );

      buf += tmp;
      count -= tmp;
      result += tmp;

      CLI( &flags );
      pchn -> frag_size -= tmp;
      if ( !pchn -> frag_size )
        {
          pchn -> used--;
          pchn -> tail++;
          pchn -> tail %= pchn -> blocks;
          pchn -> frag_size = pchn -> block_size;
        }
      STI( &flags );
    }

  return result;
}

/*
 *  synchronize playback
 */

static void ultra_pcm_sync_playback( ultra_pcm_t *pcm )
{
  unsigned long flags;
  ultra_pcm_channel_t *pchn;

  pchn = &pcm -> playback;
  if ( pchn -> flags & ULTRA_PCM_FLG_ABORT ) {
    if ( pchn -> flags & ULTRA_PCM_FLG_TRIGGER ) {
      pchn -> hw.trigger( pcm, 0 );
      pchn -> flags &= ~ULTRA_PCM_FLG_TRIGGER;
    }
    goto __end_done;
  }
  pchn -> flags |= ULTRA_PCM_FLG_SYNC;
  if ( pchn -> used < pchn -> blocks && pchn -> frag_size > 0 )
    {
      unsigned int size = pchn -> frag_size;
      unsigned int count = pchn -> block_size - size;
      unsigned int offset;;
      
      if ( count > 0 ) {
        offset = ( pchn -> head * pchn -> block_size ) + size;
        pchn -> hw.dma_neutral( pcm, pchn -> buffer, offset, count, pchn -> neutral_byte );
      }
      pchn -> used++;
      pchn -> head++;
      pchn -> head %= pchn -> blocks;
    }
  if ( !pchn -> used ) goto __end;	/* probably mmaped access */

  ultra_pcm_trigger_playback( pcm, pchn );

  while ( pchn -> used ) {
    CLI( &flags );
    pchn -> flags |= ULTRA_PCM_FLG_SLEEP;
    SLEEP( pcm, playback, 10 * HZ );
    pchn -> flags &= ~ULTRA_PCM_FLG_SLEEP;
    STI( &flags );
    if ( TABORT( pcm, playback ) ) {
      pchn -> flags |= ULTRA_PCM_FLG_ABORT;
      pchn -> hw.trigger( pcm, 0 );
      pchn -> flags &= ~ULTRA_PCM_FLG_TRIGGER;
      goto __end_done;
    }
    if ( pchn -> used && TIMEOUT( pcm, playback ) ) {
      PRINTD( "ultra_pcm_sync_playback: timeout, skipping waiting blocks\n" );
      return;
    }
  }
  
  __end_done:
  pchn -> flags |= ULTRA_PCM_FLG_NEUTRAL;
  ultra_pcm_fill_with_neutral( pchn );
  pchn -> flags &= ~ULTRA_PCM_FLG_TRIGGER1;
  pchn -> head = pchn -> tail = pchn -> used = 0;
  ULTRA_PCM_LOCKZERO( pchn );
  __end:
  pchn -> flags &= ~ULTRA_PCM_FLG_SYNC;
}

static void ultra_pcm_sync_record( ultra_pcm_t *pcm )
{
  unsigned long flags;
  ultra_pcm_channel_t *pchn;

  pchn = &pcm -> record;
  pchn -> flags |= ULTRA_PCM_FLG_SYNC;
  while ( pchn -> flags & ULTRA_PCM_FLG_TRIGGER ) {	/* still recording? */
    CLI( &flags );
    pchn -> flags |= ULTRA_PCM_FLG_SLEEP;
    SLEEP( pcm, record, 10 * HZ );
    pchn -> flags &= ~ULTRA_PCM_FLG_SLEEP;
    STI( &flags );
    if ( TABORT( pcm, record ) ) {
      pchn -> flags |= ULTRA_PCM_FLG_ABORT;
      pchn -> hw.trigger( pcm, 0 );
      pchn -> flags &= ~(ULTRA_PCM_FLG_SYNC|ULTRA_PCM_FLG_TRIGGER);
      return;
    }
    if ( (pchn -> flags & ULTRA_PCM_FLG_TRIGGER) && TIMEOUT( pcm, record ) )
      { PRINTD( "ultra_pcm_sync_record: sync timeout\n" );
      pchn -> flags &= ~ULTRA_PCM_FLG_SYNC;
      return;
    }    
  }
  pchn -> flags &= ~ULTRA_PCM_FLG_SYNC;
  ultra_pcm_fill_with_neutral( pchn );
  pchn -> flags &= ~ULTRA_PCM_FLG_TRIGGER1;
  pchn -> head = pchn -> tail = pchn -> used = 0;
  ULTRA_PCM_LOCKZERO( pchn );
}

static int ultra_pcm_sync( ultra_pcm_t *pcm, unsigned short flags )
{
  if ( flags & ULTRA_PCM_LFLG_PLAY ) ultra_pcm_sync_playback( pcm );
  if ( flags & ULTRA_PCM_LFLG_RECORD ) ultra_pcm_sync_record( pcm );
  return 0;
}

/*
 *  other things
 */

static void ultra_pcm_fill_with_neutral( ultra_pcm_channel_t *pchn )
{
  MEMSET( pchn -> buffer, pchn -> neutral_byte, pchn -> used_size );
  pchn -> flags &= ~ULTRA_PCM_FLG_NEUTRAL;
}

static void ultra_pcm_compute_blocks( ultra_pcm_t *pcm, ultra_pcm_channel_t *pchn )
{
  int record;
  unsigned int block, blocks, bps;
  unsigned long flags;
#if 0
  unsigned int max;
#endif

  record = pchn == &pcm -> record;
  if ( !record )
    ultra_pcm_sync_playback( pcm );
   else
    ultra_pcm_sync_record( pcm );
  CLI( &flags );
  if ( pchn -> requested_block_size )
    {
      block = pchn -> requested_block_size;
#if 0
      if ( block > pchn -> size / pchn -> requested_blocks )
        block = pchn -> size / pchn -> requested_blocks;
#endif
      block &= ~3;	/* align to 4 */
      if ( pcm -> info_flags & ULTRA_PCM_INFO_PLAYBACK_BATCH )	/* GF1? */
        block &= ~31;
      blocks = pchn -> size / block;
      if ( blocks > pchn -> requested_blocks )
        blocks = pchn -> requested_blocks;
    }
   else
    {
      bps = pchn -> rate * pchn -> voices;
      if ( pchn -> mode & ULTRA_PCM_MODE_16 ) bps <<= 1;
      if ( pchn -> mode & ULTRA_PCM_MODE_ADPCM ) bps >>= 2;
      block = pchn -> size;
      while ( block > bps ) block >>= 1;
      if ( block == pchn -> size ) block >>= 1;
      if ( record )
        block /= 4;			/* small fragment when recording */
      block &= ~3;			/* align to 4 */
      blocks = pchn -> size / block;
    }
#if 0
   max = pchn -> hw_max_dma_block( card );
   if ( max < block ) block = max;
#endif
   if ( blocks > 128 ) blocks = 128;
   pchn -> used_size = blocks * block;
   pchn -> blocks = blocks;
   pchn -> block_size = block;
   pchn -> flags |= ULTRA_PCM_FLG_NEUTRAL;
   pchn -> flags &= ~(ULTRA_PCM_FLG_BUFFERS|ULTRA_PCM_FLG_REALTM);
   if ( record )
     pchn -> frag_size = pchn -> block_size;
   STI( &flags );
#ifdef ULTRA_PCM_DEBUG_BUFFERS
   PRINTK( "used_size = %i, blocks = %i, blocks_size = %i, mmap = %s\n", pchn -> used_size, pchn -> blocks, pchn -> block_size, pchn -> flags & ULTRA_PCM_FLG_MMAP ? "yes" : "no" );
#endif
}

static int ultra_pcm_set_subdivision( ultra_pcm_t *pcm, unsigned short flags, unsigned int subdivision )
{
  int idx;
  ultra_pcm_channel_t *pchn;

  for ( idx = 1; idx <= 2; idx++ )
    if ( flags & idx )
      {
        pchn = idx == 1 ? &pcm -> playback : &pcm -> record;
        if ( pchn -> requested_subdivision != subdivision )
          {
            if ( idx == 1 )
              ultra_pcm_sync_playback( pcm );
             else
              ultra_pcm_sync_record( pcm );
            pchn -> requested_subdivision = subdivision;
            pchn -> flags |= ULTRA_PCM_FLG_NEUTRAL | ULTRA_PCM_FLG_BUFFERS;
          }
      }  
  return 0;
}

/*
 * SET FRAGMENT notes:
 *   Real Video 5.0b3 - (1) fragment = 0x2e00004 (rate & format set before)
 *                      (2) fragment = 0x04 (rate & format set before)
 *   Libmmoss 1.1 - fragment = 0x20009 (rate & format set before)
 *   MTV - fragment = 0x7f000a (rate & format set after!!)
 */

static int ultra_pcm_set_fragment( ultra_pcm_t *pcm, unsigned short flags, unsigned int fragment )
{
  int idx, size, size1, bytes, count, min_fragment;
  ultra_pcm_channel_t *pchn;

  size = 0x10000;
  count = 128;
  bytes = fragment & 0xffff;
  if ( !bytes ) {
    if ( flags & ULTRA_PCM_LFLG_PLAY )
      size = pcm -> playback.requested_block_size;
    if ( flags & ULTRA_PCM_LFLG_RECORD )
      size = pcm -> record.requested_block_size;
    if ( !size ) size = 0x10000;
    size1 = size;
    while ( size1 > 0 ) { bytes++; size1 >>= 1; }
  } else {
    min_fragment = 4;
    if ( flags & ULTRA_PCM_LFLG_PLAY ) 
      if ( pcm -> playback.hw.min_fragment > min_fragment )
        min_fragment = pcm -> playback.hw.min_fragment;
    if ( flags & ULTRA_PCM_LFLG_RECORD ) 
      if ( pcm -> record.hw.min_fragment > min_fragment )
        min_fragment = pcm -> record.hw.min_fragment;
    if ( bytes < min_fragment ) bytes = min_fragment;
    if ( bytes > 17 ) bytes = 17;
    size = 1 << bytes;
  }
  if ( !count ) {
    count = 128;
  } else {
    if ( count < 2 ) count = 2;
    if ( count > 128 ) count = 128;
  }

#ifdef ULTRA_PCM_DEBUG_BUFFERS
  printk( "set fragment: fragment = 0x%x, size = %i, count = %i\n", fragment, size, count );
#endif

  for ( idx = 1; idx <= 2; idx++ )
    if ( flags & idx )
      {
        pchn = idx == 1 ? &pcm -> playback : &pcm -> record;
        if ( pchn -> requested_block_size != size ||
             pchn -> requested_blocks != count )
          {
            if ( idx == 1 )
              ultra_pcm_sync_playback( pcm );
             else
              ultra_pcm_sync_record( pcm );
            pchn -> requested_block_size = size;
            pchn -> requested_blocks = count;
            pchn -> flags |= ULTRA_PCM_FLG_NEUTRAL | ULTRA_PCM_FLG_BUFFERS;
          }
      }  
  return ( count << 16 ) | bytes;
}


static unsigned int ultra_pcm_get_block_size( ultra_pcm_t *pcm, unsigned short flags )
{
  ultra_card_t *card;

  card = pcm -> card;
  if ( ( flags & ULTRA_PCM_LFLG_PLAY ) && ( pcm -> playback.flags & ULTRA_PCM_FLG_BUFFERS ) )
    ultra_pcm_compute_blocks( pcm, &pcm -> playback );
  if ( ( flags & ULTRA_PCM_LFLG_RECORD ) && ( pcm -> record.flags & ULTRA_PCM_FLG_BUFFERS ) )
    ultra_pcm_compute_blocks( pcm, &pcm -> record );
  return (flags & 1) ? pcm -> playback.block_size : pcm -> record.block_size;
}

static void ultra_pcm_set_mode( ultra_pcm_t *pcm, unsigned short flags, unsigned int mode )
{
  int idx;
  ultra_pcm_channel_t *pchn;

#ifdef ULTRA_PCM_DEBUG_BUFFERS
  printk( "mode = 0x%x\n", mode );
#endif
  for ( idx = 1; idx <= 2; idx++ )
    if ( flags & idx )
      {
        pchn = idx == 1 ? &pcm -> playback : &pcm -> record;
        if ( pchn -> mode != mode )
          {
            if ( idx == 1 )
              ultra_pcm_sync_playback( pcm );
             else 
              ultra_pcm_sync_record( pcm );
            pchn -> neutral_byte = ( mode & ULTRA_PCM_MODE_16 ) ? 0x00 : 0x80;
            pchn -> mode = mode;
            pchn -> flags |= ULTRA_PCM_FLG_NEUTRAL | ULTRA_PCM_FLG_BUFFERS;
          }
      }
}

static unsigned int ultra_pcm_get_mode( ultra_pcm_t *pcm, unsigned short flags )
{
  return (flags & 1) ? pcm -> playback.mode : pcm -> record.mode;
}

static void ultra_pcm_set_format( ultra_pcm_t *pcm, unsigned short flags, unsigned int format )
{
  int idx;
  ultra_pcm_channel_t *pchn;

#ifdef ULTRA_PCM_DEBUG_BUFFERS
  printk( "format = 0x%x\n", format );
#endif
  for ( idx = 1; idx <= 2; idx++ )
    if ( flags & idx )
      {
        pchn = idx == 1 ? &pcm -> playback : &pcm -> record;
        pchn -> format = format;
      }
}

static unsigned int ultra_pcm_get_format( ultra_pcm_t *pcm, unsigned short flags )
{
  return (flags & 1) ? pcm -> playback.format : pcm -> record.format;
}

static void ultra_pcm_set_rate( ultra_pcm_t *pcm, unsigned short flags, unsigned int rate )
{
  int idx;
  ultra_pcm_channel_t *pchn;

#ifdef ULTRA_PCM_DEBUG_BUFFERS
  printk( "rate = %i\n", rate );
#endif
  for ( idx = 1; idx <= 2; idx++ )
    if ( flags & idx )
      {
        pchn = idx == 1 ? &pcm -> playback : &pcm -> record;
        if ( pchn -> rate != rate )
          {
            if ( idx == 1 )
              ultra_pcm_sync_playback( pcm );
             else
              ultra_pcm_sync_record( pcm );
            pchn -> rate = rate;
            pchn -> flags |= ULTRA_PCM_FLG_BUFFERS; 
          }
    }
}

static unsigned int ultra_pcm_get_rate( ultra_pcm_t *pcm, unsigned short flags )
{
  return (flags & 1) ? pcm -> playback.rate : pcm -> record.rate;
}

static int ultra_set_format( ultra_pcm_t *pcm, unsigned short flags, int format )
{
  unsigned int new_mode;
  unsigned int new_format;

  if ( format != ULTRA_PCM_FMT_QUERY )
    {
#if 0
      printk( "format = 0x%x\n", format );
#endif
      if ( !format )
        {
          new_format = ultra_pcm_get_format( pcm, flags );
          new_mode = ultra_pcm_get_mode( pcm, flags );
        }
       else
        {
          new_format = 0;
          if ( flags & ULTRA_PCM_LFLG_PLAY ) {
            new_format = format & pcm -> playback.hw.formats;
            if ( !new_format )			/* not supported format */
              new_format = ULTRA_PCM_FMT_U8;	/* always supported */
          }
          if ( flags & ULTRA_PCM_LFLG_RECORD ) {
            if ( new_format )
              new_format &= pcm -> record.hw.formats;
             else
              new_format = format & pcm -> record.hw.formats;
            if ( !new_format )			/* not supported format */
              new_format = ULTRA_PCM_FMT_U8;	/* always supported */
          }
          new_mode = ultra_pcm_get_mode( pcm, flags ) & ~ULTRA_PCM_MODE_TYPE;
          new_mode |= ULTRA_PCM_MODE_VALID;
          if ( new_format & ( ULTRA_PCM_FMT_MU_LAW | ULTRA_PCM_FMT_A_LAW | ULTRA_PCM_FMT_U8 | 
                              ULTRA_PCM_FMT_U16_LE | ULTRA_PCM_FMT_U16_BE ) )
            new_mode |= ULTRA_PCM_MODE_U;
          if ( new_format & ( ULTRA_PCM_FMT_S16_LE | ULTRA_PCM_FMT_S16_BE | ULTRA_PCM_FMT_U16_LE | 
                              ULTRA_PCM_FMT_U16_BE | ULTRA_PCM_FMT_IMA_ADPCM ) )
            new_mode |= ULTRA_PCM_MODE_16;
          if ( new_format & ULTRA_PCM_FMT_MU_LAW )
            new_mode |= ULTRA_PCM_MODE_ULAW;
          if ( new_format & ULTRA_PCM_FMT_A_LAW )
            new_mode |= ULTRA_PCM_MODE_ALAW;
          if ( new_format & ULTRA_PCM_FMT_IMA_ADPCM )
            new_mode |= ULTRA_PCM_MODE_ADPCM;
        }
      if ( new_mode != ultra_pcm_get_mode( pcm, flags ) )
        {
          ultra_pcm_set_format( pcm, flags, new_format );
          ultra_pcm_set_mode( pcm, flags, new_mode );
        }
    }

  return ultra_pcm_get_format( pcm, flags );
}

static int ultra_set_rate( ultra_pcm_t *pcm, unsigned short flags, unsigned int rate )
{
  unsigned int max_rate;

  if ( rate > 0 ) 
    {
      max_rate = 48000;
      if ( flags & ULTRA_PCM_LFLG_PLAY ) {
        if ( max_rate > pcm -> playback.hw.max_rate )
          max_rate = pcm -> playback.hw.max_rate;
      }
      if ( flags & ULTRA_PCM_LFLG_RECORD ) {
        if ( max_rate > pcm -> record.hw.max_rate )
          max_rate = pcm -> record.hw.max_rate;
      }
      if ( rate > max_rate ) rate = max_rate;
      ultra_pcm_set_rate( pcm, flags, rate );
    }
   else
    rate = ultra_pcm_get_rate( pcm, flags );

  return rate;
}

static int ultra_pcm_set_channels( ultra_pcm_t *pcm, unsigned short flags, int channels )
{
  ultra_pcm_channel_t *pchn;

#if 0
  printk( "set channels = %i\n", channels );
#endif
  if ( channels >= 0 )
    {
      if ( channels < 1 || channels > 32 ) return -EINVAL;
      if ( flags & ULTRA_PCM_LFLG_PLAY ) {
        pchn = &pcm -> playback;
        if ( channels > pchn -> hw.max_voices ) return -EINVAL;
        ultra_pcm_sync_playback( pcm );
        pchn -> voices = channels;
        pchn -> flags |= ULTRA_PCM_FLG_BUFFERS;
      }
      if ( flags & ULTRA_PCM_LFLG_RECORD ) {
        pchn = &pcm -> record;
        if ( channels > pchn -> hw.max_voices ) return -EINVAL;
        ultra_pcm_sync_playback( pcm );
        pchn -> voices = channels;
        pchn -> flags |= ULTRA_PCM_FLG_BUFFERS;
      }
    }
   else
    {
      if ( flags & ULTRA_PCM_LFLG_PLAY )
        channels = pcm -> playback.voices;
      if ( flags & ULTRA_PCM_LFLG_RECORD )
        channels = pcm -> playback.voices;
    }
  return channels;
}

/*
 * This is probably wrong implementation.
 * Now is better way trigger out appropriate channel
 * and reset head, tail & used variables.
 */

static int ultra_pcm_reset( ultra_pcm_t *pcm, unsigned short flags )
{
  int idx;
  ultra_pcm_channel_t *pchn;

#ifdef ULTRA_PCM_DEBUG_BUFFERS
  printk( "pcm_reset!!!\n" );
#endif
  for ( idx = 1; idx <= 2; idx++ )
    if ( flags & idx )
      {
        pchn = idx == 1 ? &pcm -> playback : &pcm -> record;
        if ( idx == 1 )
          ultra_pcm_sync_playback( pcm );
         else
          ultra_pcm_sync_record( pcm );
#if 0
        pchn -> requested_block_size =
        pchn -> requested_blocks =
        pchn -> requested_subdivision = 0;
        pchn -> processed_bytes = 0;
        pchn -> interrupts = 0;
        pchn -> flags |= ULTRA_PCM_FLG_BUFFERS;
#endif
        pchn -> flags &= ~ULTRA_PCM_FLG_ABORT;
      }
  return 0;
}

static int ultra_pcm_get_trigger( ultra_pcm_t *pcm, unsigned short flags )
{
  unsigned long iflags;
  int result;

  result = 0;
  CLI( &iflags );
  if ( ( pcm -> playback.flags & ULTRA_PCM_FLG_TRIGGER ) && ( flags & ULTRA_PCM_LFLG_PLAY ) ) result |= ULTRA_PCM_ENABLE_PLAYBACK;
  if ( ( pcm -> record.flags & ULTRA_PCM_FLG_TRIGGER ) && ( flags & ULTRA_PCM_LFLG_RECORD ) ) result |= ULTRA_PCM_ENABLE_RECORD;
  STI( &iflags );
#if 0
  printk( "get trigger = 0x%x\n", result );
#endif
  return result;
}

static int ultra_pcm_set_trigger( ultra_pcm_t *pcm, unsigned short flags, int trigger )
{
  ultra_card_t *card;
  ultra_pcm_channel_t *pchn;

#if 0
  printk( "set trigger = 0x%x\n", trigger );
#endif
  card = pcm -> card;
  if ( flags & ULTRA_PCM_LFLG_PLAY ) {
    pchn = &pcm -> playback;
    if ( trigger & ULTRA_PCM_ENABLE_PLAYBACK ) {
      pchn -> flags |= ULTRA_PCM_FLG_ENABLE;
      if ( !(pchn -> flags & ULTRA_PCM_FLG_TRIGGER) ) {
        if ( pchn -> flags & ULTRA_PCM_FLG_BUFFERS )
          ultra_pcm_compute_blocks( pcm, pchn );
        if ( pchn -> flags & ULTRA_PCM_FLG_MMAP ) {
          if ( !(pchn -> flags & ULTRA_PCM_FLG_TRIGGER1) ) {	/* first trigger? */
            pchn -> hw.prepare( pcm,
                        	pchn -> buffer, pchn -> used_size,
            			0,				/* assuming that first block is always 0 */
            			pchn -> block_size );
          }
          pchn -> hw.trigger( pcm, 1 );
          pchn -> flags |= ULTRA_PCM_FLG_TRIGGERA;
        } else {
          ultra_pcm_trigger_playback( pcm, pchn );
        }
      }
    } else {
      pchn -> flags &= ~ULTRA_PCM_FLG_ENABLE;
      if ( pchn -> flags & ULTRA_PCM_FLG_TRIGGER ) {
        pchn -> hw.trigger( pcm, 0 );
        pchn -> flags &= ~ULTRA_PCM_FLG_TRIGGER;
      }
    }
  } 
  if ( flags & ULTRA_PCM_LFLG_RECORD ) {
    pchn = &pcm -> record;
    if ( trigger & ULTRA_PCM_ENABLE_RECORD ) {
      pchn -> flags |= ULTRA_PCM_FLG_ENABLE;
      if ( !(pchn -> flags & ULTRA_PCM_FLG_TRIGGER) ) {
        if ( pchn -> flags & ULTRA_PCM_FLG_BUFFERS )
          ultra_pcm_compute_blocks( pcm, pchn );
        if ( pchn -> flags & ULTRA_PCM_FLG_MMAP ) {
          if ( !(pchn -> flags & ULTRA_PCM_FLG_TRIGGER1) ) {	/* first trigger? */
            pchn -> hw.prepare( pcm,
            			pchn -> buffer, pchn -> used_size,
            			0,
            			pchn -> block_size );
          }
          pchn -> hw.trigger( pcm, 1 );
          pchn -> flags |= ULTRA_PCM_FLG_TRIGGERA;
        } else {
          ultra_pcm_trigger_record( pcm, pchn );
        }
      }
    } else {
      pchn -> flags &= ~ULTRA_PCM_FLG_ENABLE;
      if ( pchn -> flags & ULTRA_PCM_FLG_TRIGGER ) {
        pchn -> hw.trigger( pcm, 0 );
        pchn -> flags &= ~ULTRA_PCM_FLG_TRIGGER;
      }
    }
  }
  return ultra_pcm_get_trigger( pcm, flags );
}

#ifdef ULTRACFG_OSS

static int ultra_pcm_trigger_sequencer( unsigned int devmask )
{
  int i;
  ultra_card_t *card;
  ultra_pcm_t *pcm;
  
  for ( i = 0; i < ultra_cards_count * 2; i++ ) {
    if ( ( devmask & (1 << i) ) == 0 ) continue; 
    card = ultra_cards[ i >> 1 ];
    pcm = &card -> pcm[ i & 1 ];
    if ( !(pcm -> flags & ULTRA_PCM_LFLG_PLAY) ) continue;
    pcm_set_trigger( pcm, ULTRA_PCM_LFLG_PLAY, ULTRA_PCM_ENABLE_OUTPUT );
  }
  return 0;
}

#endif

static int ultra_pcm_get_space( ultra_pcm_t *pcm, unsigned short flags, int record, struct ULTRA_PCM_BUFFER_INFO *arg )
{
  unsigned long iflags;
  ultra_pcm_channel_t *pchn;
  struct ULTRA_PCM_BUFFER_INFO info;

  if ( !( flags & ( 1 << record ) ) ) return -EIO;	/* bad file mode */
  if ( VERIFY_AREA( VERIFY_WRITE, arg, sizeof( info ) ) ) return -EIO;
  pchn = !record ? &pcm -> playback : &pcm -> record;
  if ( pchn -> flags & ULTRA_PCM_FLG_BUFFERS )
    ultra_pcm_compute_blocks( pcm, pchn );
  CLI( &iflags );
  info.fragments = pchn -> blocks - pchn -> used;
  info.fragstotal = pchn -> blocks;
  info.fragsize = pchn -> block_size;
  info.bytes = info.fragments * info.fragsize;
  if ( (pchn -> flags & ULTRA_PCM_FLG_TRIGGER) || pchn -> used )
    info.bytes -= (pchn -> used * pchn -> block_size) + pchn -> used_size;
  STI( &iflags );
#ifdef ULTRA_PCM_DEBUG_BUFFERS
  printk( "space: frags = %i, total = %i, size = %i, bytes = %i\n", info.fragments, info.fragstotal, info.fragsize, info.bytes );
#endif
  pchn -> flags |= ULTRA_PCM_FLG_ENABLE;
  MEMCPY_TOFS( arg, &info, sizeof( info ) );
  return 0;
}

static int ultra_pcm_get_ptr( ultra_pcm_t *pcm, unsigned short flags, int record, struct ULTRA_PCM_COUNT_INFO *arg )
{
  unsigned long iflags;
  ultra_pcm_channel_t *pchn;
  struct ULTRA_PCM_COUNT_INFO info;

  if ( !( flags & ( 1 << record ) ) ) return -EIO;	/* bad file mode */
  if ( VERIFY_AREA( VERIFY_WRITE, arg, sizeof( info ) ) ) return -EIO;
  pchn = !record ? &pcm -> playback : &pcm -> record;
  if ( pchn -> flags & ULTRA_PCM_FLG_BUFFERS )
    ultra_pcm_compute_blocks( pcm, pchn );
  CLI( &iflags );
  pchn -> flags |= ULTRA_PCM_FLG_REALTM;
  info.bytes = pchn -> processed_bytes;
  if ( !(pchn -> flags & ULTRA_PCM_FLG_TRIGGER) )
    info.ptr = pchn -> used_size;
   else {
    info.ptr = pchn -> hw.pointer( pcm, pchn -> used_size );
    if ( info.ptr ) info.ptr--;
  }
  info.blocks = pchn -> used;
  if ( pchn -> flags & ULTRA_PCM_FLG_MMAP ) {
    info.bytes += info.ptr;
    info.blocks = pchn -> interrupts;
    pchn -> interrupts = 0;
  } else {
    if ( !info.bytes )
      info.bytes = info.ptr = pchn -> size;
    info.bytes += info.ptr % pchn -> block_size;
    info.blocks = pchn -> used;
  }
#if 0
  printk( "ptr: flags = 0x%x, bytes = %i, ptr = %i, blocks = %i\n", pchn -> flags, info.bytes, info.ptr, info.blocks );
#endif
  STI( &iflags );
  MEMCPY_TOFS( arg, &info, sizeof( info ) );
  return 0;
}

static int ultra_pcm_get_mapbuf( ultra_pcm_t *pcm, unsigned short flags, int record, struct ULTRA_PCM_BUFFER_DESCRIPTION *arg )
{
  unsigned long iflags;
  ultra_pcm_channel_t *pchn;
  struct ULTRA_PCM_BUFFER_DESCRIPTION info;

  if ( !( flags & ( 1 << record ) ) ) return -EIO;	/* bad file mode */
  if ( VERIFY_AREA( VERIFY_WRITE, arg, sizeof( info ) ) ) return -EIO;
  pchn = !record ? &pcm -> playback : &pcm -> record;
  CLI( &iflags );
  info.buffer = (unsigned char *)pchn -> buffer;
  info.size = pchn -> size;
  STI( &iflags );
  MEMCPY_TOFS( arg, &info, sizeof( info ) );
  return 0;
}

static int ultra_pcm_info_flags( ultra_pcm_t *pcm, struct ULTRA_PCM_INFO *info, int control )
{
  struct ULTRA_PCM_INFO ginfo;
  ultra_pcm_channel_t *pchn_p, *pchn_r;

  if ( control ) {
    int card, device;
    
    if ( VERIFY_AREA( VERIFY_READ, info, sizeof( ginfo ) ) ) return -EIO;
    MEMCPY_FROMFS( &ginfo, info, sizeof( ginfo ) );
    card = ginfo.device >> 16;
    if ( card >= ultra_cards_count ) return -ENXIO;
    device = ginfo.device & 0xffff;
    if ( device >= ULTRA_PCM_DEVICES ) return -ENXIO;
    pcm = ultra_pcm_devices[ (card * ULTRA_PCM_DEVICES) + device ];
    if ( !pcm ) return -ENODEV;
    if ( !(pcm -> flags & ULTRA_PCM_LFLG_USED) ) return -ENXIO;
  }
  if ( VERIFY_AREA( VERIFY_WRITE, info, sizeof( ginfo ) ) ) return -EIO;
  MEMSET( &ginfo, 0, sizeof( ginfo ) );
  pchn_p = &pcm -> playback;
  pchn_r = &pcm -> record;
  ginfo.flags = pcm -> info_flags;
  strcpy( ginfo.name, pcm -> info_name );
  ginfo.max_rate_play = pchn_p -> hw.max_rate;
  ginfo.max_rate_record = pchn_r -> hw.max_rate;
  ginfo.formats_play = pchn_p -> hw.formats;
  ginfo.formats_record = pchn_r -> hw.formats;
  ginfo.max_channels_play = pchn_p -> hw.max_voices;
  ginfo.max_channels_record = pchn_r -> hw.max_voices;
  ginfo.dma_size_play = pchn_p -> size;
  ginfo.dma_size_record = pchn_r -> size;
  MEMCPY_TOFS( info, &ginfo, sizeof( ginfo ) );
  return 0;
}

static int ultra_get_osscaps( ultra_pcm_t *pcm, int flags )
{
  unsigned int result;
  
  result = /* ULTRA_PCM_CAP_REALTIME | */ 
           ULTRA_PCM_CAP_TRIGGER | ULTRA_PCM_CAP_MMAP;
  if ( flags & ULTRA_PCM_LFLG_PLAY ) {
    if ( pcm -> info_flags & ULTRA_PCM_INFO_PLAYBACK_BATCH )
      result |= ULTRA_PCM_CAP_BATCH;
  } else {
    if ( pcm -> info_flags & ULTRA_PCM_INFO_RECORD_BATCH )
      result |= ULTRA_PCM_CAP_BATCH;
  }
  if ( (pcm -> info_flags & ULTRA_PCM_INFO_DUPLEX) && flags == ULTRA_PCM_LFLG_BOTH )
    result |= ULTRA_PCM_CAP_DUPLEX;
  result |= 0x0001;	/* revision - same as SB AWE 64 */
  return result;
}

static int ultra_pcm_nonblock( ultra_pcm_t *pcm, int flags )
{
  if ( flags & ULTRA_PCM_LFLG_PLAY )
    pcm -> playback.flags |= ULTRA_PCM_FLG_NONBLK;
  if ( flags & ULTRA_PCM_LFLG_RECORD )
    pcm -> record.flags |= ULTRA_PCM_FLG_NONBLK;
  return 0;
}

static int ultra_pcm_open_card( ultra_pcm_t *pcm, short minor, struct file *file )
{
  unsigned short flags;
  int res;

  flags = ultra_pcm_file_flags( file -> f_flags );
#ifdef ULTRA_PCM_DEBUG_BUFFERS
  PRINTK( "ultra_pcm_open (1) - pcm = 0x%lx, minor = %i, flags = %i, pcm->flags = 0x%x\n", (long)pcm, minor, flags, pcm -> flags );
#endif
  if ( flags & pcm -> flags ) return -EBUSY;	/* channel(s) already used */
  if ( flags & ULTRA_PCM_LFLG_PLAY )
    if ( !(pcm -> info_flags & ULTRA_PCM_INFO_PLAYBACK) ) return -ENODEV;
  if ( flags & ULTRA_PCM_LFLG_RECORD )
    if ( !(pcm -> info_flags & ULTRA_PCM_INFO_RECORD) ) return -ENODEV;
  if ( flags == ULTRA_PCM_LFLG_BOTH )
    if ( !(pcm -> info_flags & ULTRA_PCM_INFO_DUPLEX) ) return -ENODEV;
  if ( flags & ULTRA_PCM_LFLG_PLAY )
    ultra_pcm_clear_channel( &pcm -> playback );
  if ( flags & ULTRA_PCM_LFLG_RECORD )
    ultra_pcm_clear_channel( &pcm -> record );
  pcm -> mask = ULTRA_PCM_LFLG_BOTH;
  ultra_pcm_set_format( pcm, flags, 0 );
  ultra_pcm_set_mode( pcm, flags, 0 );
  if ( minor < ULTRA_MINOR_UBEGIN ) {
    switch ( minor & ULTRA_MINOR_OSS_MASK ) {
      case ULTRA_MINOR_AUDIO:
        ultra_set_format( pcm, flags, ULTRA_PCM_FMT_MU_LAW );
        break;
      case ULTRA_MINOR_PCM_8:
        ultra_set_format( pcm, flags, ULTRA_PCM_FMT_U8 );
        break; 
      case ULTRA_MINOR_PCM_16:
      case ULTRA_MINOR_PCM1:
        ultra_set_format( pcm, flags, ULTRA_PCM_FMT_U16_LE );
        break;
      default:
        PRINTD( "pcm: bad minor value\n" );
        return -EINVAL;
    }
  } else {
    ultra_set_format( pcm, flags, ULTRA_PCM_FMT_U16_LE );
  }
  ultra_pcm_set_rate( pcm, flags, ULTRA_PCM_DEFAULT_RATE );
  ultra_pcm_set_channels( pcm, flags, 1 );
  if ( flags & ULTRA_PCM_LFLG_PLAY )
    {
      if ( ( res = pcm -> playback.hw.open( pcm ) ) < 0 )
        return res;
      pcm -> playback.flags |= ULTRA_PCM_FLG_ENABLE;
    }
  if ( flags & ULTRA_PCM_LFLG_RECORD )
    {
      if ( ( res = pcm -> record.hw.open( pcm ) ) < 0 )
        {
          if ( flags & ULTRA_PCM_LFLG_PLAY )
            {
              pcm -> playback.hw.close( pcm );
              pcm -> playback.flags &= ~ULTRA_PCM_FLG_ENABLE;
            }
          return res;
        }
      pcm -> record.flags |= ULTRA_PCM_FLG_ENABLE;
    }
  pcm -> flags |= flags;

#if 0
  printk( "pcm open - done...\n" );
#endif

  return 0;
}

static void ultra_pcm_close_card( ultra_pcm_t *pcm, struct file *file )
{
  unsigned short flags;
  
  if ( !pcm ) return;
  flags = ultra_pcm_file_flags( file -> f_flags ) & pcm -> flags;
#if 0
  PRINTK( "ultra_release_pcm - flags = %i\n", flags );
#endif
  if ( flags & ULTRA_PCM_LFLG_PLAY )
    {
      ultra_pcm_sync_playback( pcm );		/* synchronize playback */
      pcm -> playback.hw.close( pcm );
      pcm -> flags &= ~ULTRA_PCM_LFLG_PLAY;
    }
  if ( flags & ULTRA_PCM_LFLG_RECORD )
    {
      ultra_pcm_sync_record( pcm );		/* and record */
      pcm -> record.hw.close( pcm );
      pcm -> flags &= ~ULTRA_PCM_LFLG_RECORD;
    }
#if 0
  printk( "release pcm: done\n" );
#endif
}

static int ultra_pcm_open( unsigned short minor, struct file *file )
{
  short dev;
  int idx, res;
  ultra_pcm_t *pcm;

  dev = ( minor >> 4 ) & 7;
  if ( dev >= ultra_cards_count ) return -ENODEV;
  idx = 0;
  if ( minor >= ULTRA_MINOR_UBEGIN ) {
    if ( ( minor & ULTRA_MINOR_UDEVMASK ) == ULTRA_MINOR_UPCM1 ) idx = 1;
  } else {
    if ( ( minor & ULTRA_MINOR_OSS_MASK ) == ULTRA_MINOR_PCM1 ) idx = 1;
  }
  file -> private_data = pcm = ultra_pcm_devices[ (dev * ULTRA_PCM_DEVICES) + idx ];
  if ( ( res = ultra_pcm_open_card( pcm, minor, file ) ) < 0 )
    {
      file -> private_data = NULL;
      return res;
    }
  MOD_INC_USE_COUNT;
  pcm -> card -> use_inc();
  return 0;
}

static int ultra_pcm_release( unsigned short minor, struct file *file )
{
  ultra_pcm_t *pcm;
 
  if ( file -> private_data ) {
    pcm = (ultra_pcm_t *)file -> private_data;
    ultra_pcm_close_card( pcm, file );
    pcm -> card -> use_dec();
  }
  MOD_DEC_USE_COUNT;
  return 0;
}

static int ultra_pcm_ioctl( struct file *file, unsigned int cmd, unsigned long arg )
{
  ultra_pcm_t *pcm;
  ultra_card_t *card;
  unsigned short flags;

  pcm = (ultra_pcm_t *)file -> private_data;
  if ( !(pcm -> flags & ULTRA_PCM_LFLG_VALID) ) return -ENODEV;
  card = pcm -> card;
  if ( ( ( cmd >> 8 ) & 0xff ) == 'M' )	/* mixer ioctl - for OSS (grrr) compatibility */
    return ultra_mixer_ioctl_card( card, file, cmd, arg );
  if ( ( ( cmd >> 8 ) & 0xff ) != 'P' ) return -EIO;
  flags = ultra_pcm_file_flags( file -> f_flags ) & pcm -> mask;
#if 0
  if ( cmd != ULTRA_PCM_IOCTL_GETPBKPTR )
    printk( "cmd = 0x%x, arg = 0x%x, flags = 0x%x\n", cmd, IOCTL_IN( arg ), flags );
#endif
  switch ( cmd ) {
    case ULTRA_PCM_IOCTL_RESET:
      return ultra_pcm_reset( pcm, flags );
    case ULTRA_PCM_IOCTL_SYNC:
      return ultra_pcm_sync( pcm, flags );
    case ULTRA_PCM_IOCTL_RATE:
      return IOCTL_OUT( arg, ultra_set_rate( pcm, flags, IOCTL_IN( arg ) ) );
    case ULTRA_PCM_IOCTL_GETRATE:
      return IOCTL_OUT( arg, ultra_pcm_get_rate( pcm, flags ) );
    case ULTRA_PCM_IOCTL_STEREO:
      return IOCTL_OUT( arg, ultra_pcm_set_channels( pcm, flags, IOCTL_IN( arg ) > 0 ? 2 : 1 ) - 1 );
    case ULTRA_PCM_IOCTL_GETBLKSIZE:
      return IOCTL_OUT( arg, ultra_pcm_get_block_size( pcm, flags ) );
    case ULTRA_PCM_IOCTL_FORMAT:
      return IOCTL_OUT( arg, ultra_set_format( pcm, flags, IOCTL_IN( arg ) ) );
    case ULTRA_PCM_IOCTL_GETFORMAT:
      return IOCTL_OUT( arg, ultra_pcm_get_format( pcm, flags ) );
    case ULTRA_PCM_IOCTL_CHANNELS:
      return IOCTL_OUT( arg, ultra_pcm_set_channels( pcm, flags, IOCTL_IN( arg ) ) );
    case ULTRA_PCM_IOCTL_GETCHANNELS:
      return IOCTL_OUT( arg, ultra_pcm_set_channels( pcm, flags, -1 ) );
    case ULTRA_PCM_IOCTL_FILTER:
    case ULTRA_PCM_IOCTL_GETFILTER:
      return -EIO;
    case ULTRA_PCM_IOCTL_POST:	/* wrong implementation */
      return ultra_pcm_sync( pcm, flags );
    case ULTRA_PCM_IOCTL_SUBDIVIDE:
      return IOCTL_OUT( arg, ultra_pcm_set_subdivision( pcm, flags, IOCTL_IN( arg ) ) );
    case ULTRA_PCM_IOCTL_SETFRAGMENT:
      return IOCTL_OUT( arg, ultra_pcm_set_fragment( pcm, flags, IOCTL_IN( arg ) ) );
    case ULTRA_PCM_IOCTL_GETFORMATS:
      return IOCTL_OUT( arg, (flags & ULTRA_PCM_LFLG_PLAY) ? pcm -> playback.hw.formats : pcm -> record.hw.formats );
    case ULTRA_PCM_IOCTL_GETPBKSPACE:
    case ULTRA_PCM_IOCTL_GETRECSPACE:
      return ultra_pcm_get_space( pcm, flags, cmd == ULTRA_PCM_IOCTL_GETRECSPACE, (struct ULTRA_PCM_BUFFER_INFO *)arg );
    case ULTRA_PCM_IOCTL_NONBLOCK:
      return ultra_pcm_nonblock( pcm, flags );
    case ULTRA_PCM_IOCTL_GETCAPS:
      return IOCTL_OUT( arg, ultra_get_osscaps( pcm, flags ) );
    case ULTRA_PCM_IOCTL_GETTRIGGER:
      return IOCTL_OUT( arg, ultra_pcm_get_trigger( pcm, flags ) );
    case ULTRA_PCM_IOCTL_SETTRIGGER:
      return IOCTL_OUT( arg, ultra_pcm_set_trigger( pcm, flags, IOCTL_IN( arg ) ) );
    case ULTRA_PCM_IOCTL_GETRECPTR:
    case ULTRA_PCM_IOCTL_GETPBKPTR:
      return ultra_pcm_get_ptr( pcm, flags, cmd == ULTRA_PCM_IOCTL_GETRECPTR, (struct ULTRA_PCM_COUNT_INFO *)arg );
    case ULTRA_PCM_IOCTL_MAPRECBUFFER:
    case ULTRA_PCM_IOCTL_MAPPBKBUFFER:
      return ultra_pcm_get_mapbuf( pcm, flags, cmd == ULTRA_PCM_IOCTL_MAPRECBUFFER, (struct ULTRA_PCM_BUFFER_DESCRIPTION *)arg );
    case ULTRA_PCM_IOCTL_SYNCRO:
      /* stop DMA now.. */
      return 0;
    case ULTRA_PCM_IOCTL_DUPLEX:
      if ( ultra_get_osscaps( pcm, flags ) & ULTRA_PCM_CAP_DUPLEX ) return 0;
      return -EIO;
    case ULTRA_PCM_IOCTL_INFO:
      return ultra_pcm_info_flags( pcm, (struct ULTRA_PCM_INFO *)arg, 0 );
    case ULTRA_PCM_IOCTL_MASK:
      {
        int val;
      
        if ( ultra_pcm_file_flags( file -> f_flags ) != ULTRA_PCM_LFLG_BOTH )
          return -EIO;
        val = IOCTL_IN( arg );
        val &= ULTRA_PCM_LFLG_BOTH;
        if ( !val ) return -EINVAL;
        pcm -> mask = val;
        return 0;
      }
#ifdef ULTRACFG_DEBUG
    default:
      PRINTK( "pcm: unknown command = 0x%x\n", cmd ); 
#endif
  }
  return -EIO;
}

static long ultra_pcm_read( struct file *file, char *buf, long count )
{
  ultra_pcm_t *pcm;

  pcm = (ultra_pcm_t *)file -> private_data;
  if ( !pcm ) return -EIO;
  if ( !(pcm -> flags & ULTRA_PCM_LFLG_VALID) ) return -ENODEV;
  if ( !( ultra_pcm_file_flags( file -> f_flags ) & ULTRA_PCM_LFLG_RECORD ) ||
       !( pcm -> flags & ULTRA_PCM_LFLG_RECORD ) ) return -EIO;
  return ultra_pcm_dma_to_user( pcm, buf, count );
}

static long ultra_pcm_write( struct file *file, const char *buf, long count )
{
  ultra_pcm_t *pcm;

  pcm = (ultra_pcm_t *)file -> private_data;
  if ( !pcm ) return -EIO;
  if ( !(pcm -> flags & ULTRA_PCM_LFLG_VALID) ) return -ENODEV;
  if ( !( ultra_pcm_file_flags( file -> f_flags ) & ULTRA_PCM_LFLG_PLAY ) ||
       !( pcm -> flags & ULTRA_PCM_LFLG_PLAY ) ) return -EIO;
  return ultra_pcm_user_to_dma( pcm, buf, count );
}

#ifdef ULTRA_POLL
static unsigned int ultra_pcm_poll( struct file *file, poll_table *wait )
{
  ultra_pcm_t *pcm;
  unsigned long flags;
  unsigned int mask;
  ultra_pcm_channel_t *record, *playback;
  unsigned short fflags;

  fflags = ultra_pcm_file_flags( file -> f_flags );
  pcm = (ultra_pcm_t *)file -> private_data;
  if ( !pcm ) return 0;
  if ( !(pcm -> flags & ULTRA_PCM_LFLG_VALID) ) return 0;

  record = &pcm -> record;
  if ( (fflags & ULTRA_PCM_LFLG_RECORD) &&
       (record -> flags & ULTRA_PCM_FLG_BUFFERS) )
    ultra_pcm_compute_blocks( pcm, record );
  playback = &pcm -> playback;
  if ( (fflags & ULTRA_PCM_LFLG_PLAY) &&
       (playback -> flags & ULTRA_PCM_FLG_BUFFERS) )
    ultra_pcm_compute_blocks( pcm, playback );
  
  CLI( &flags );
  if ( fflags & ULTRA_PCM_LFLG_RECORD ) {
    record -> flags |= ULTRA_PCM_FLG_SLEEP;
    SLEEP_POLL( pcm, record, wait );
  }
  if ( fflags & ULTRA_PCM_LFLG_PLAY ) {
    playback -> flags |= ULTRA_PCM_FLG_SLEEP;
    SLEEP_POLL( pcm, playback, wait );
  }
  STI( &flags );

  mask = 0;
  if ( fflags & ULTRA_PCM_LFLG_RECORD ) {
    if ( record -> flags & ULTRA_PCM_FLG_MMAP ) {
      if ( record -> interrupts ) mask |= POLLIN | POLLRDNORM;
    } else {
      if ( record -> used ) mask |= POLLIN | POLLRDNORM;
    }
  }
  if ( fflags & ULTRA_PCM_LFLG_PLAY ) {
    if ( playback -> flags & ULTRA_PCM_FLG_MMAP ) {
      if ( playback -> interrupts ) mask |= POLLOUT | POLLWRNORM;
    } else {
      if ( playback -> used < playback -> blocks ) mask |= POLLOUT | POLLWRNORM;
    }
  }

  return mask;
}
#else
static int ultra_pcm_select( struct file *file, int sel_type, select_table *wait )
{
  int ok;
  ultra_pcm_t *pcm;
  unsigned long flags;
  ultra_pcm_channel_t *pchn;
  unsigned short fflags;

  fflags = ultra_pcm_file_flags( file -> f_flags );
  pcm = (ultra_pcm_t *)file -> private_data;
  if ( !pcm ) return -EIO;
  if ( !(pcm -> flags & ULTRA_PCM_LFLG_VALID) ) return -ENODEV;
  switch ( sel_type ) {
    case SEL_IN:
      if ( !(fflags & ULTRA_PCM_LFLG_RECORD) ) return 0;
      pchn = &pcm -> record;
      if ( pchn -> flags & ULTRA_PCM_FLG_BUFFERS )
        ultra_pcm_compute_blocks( pcm, pchn );
      CLI( &flags );
      if ( pchn -> flags & ULTRA_PCM_FLG_MMAP )
        ok = !pchn -> interrupts; 
       else
        ok = !pchn -> used;
      if ( ok )
        {
          pchn -> flags |= ULTRA_PCM_FLG_SLEEP;
          SLEEPS( pcm, record, wait );
          STI( &flags );
          return 0;
        }
      pchn -> flags &= ~ULTRA_PCM_FLG_SLEEP;
      STI( &flags );
      return 1;
    case SEL_OUT:
      if ( !(fflags & ULTRA_PCM_LFLG_PLAY) ) return 0;
      pchn = &pcm -> playback;
      if ( pchn -> flags & ULTRA_PCM_FLG_BUFFERS )
        ultra_pcm_compute_blocks( pcm, pchn );
      CLI( &flags );
      if ( pchn -> flags & ULTRA_PCM_FLG_MMAP )
        ok = !pchn -> interrupts; 
       else
        ok = pchn -> used >= pchn -> blocks;
      if ( ok )
        {
          pchn -> flags |= ULTRA_PCM_FLG_SLEEP;
          SLEEPS( pcm, playback, wait );
          STI( &flags );
          return 0;
        }
      pchn -> flags &= ~ULTRA_PCM_FLG_SLEEP;
      STI( &flags );
      return 1;
    case SEL_EX:
      break;
  }
  return 0;
}
#endif

static int ultra_pcm_mmap( struct inode *inode, struct file *file, struct vm_area_struct *vma )
{
  ultra_pcm_t *pcm;
  unsigned short flags;
  ultra_pcm_channel_t *pchn;
  unsigned long size;

  pcm = (ultra_pcm_t *)file -> private_data;
  if ( !(pcm -> flags & ULTRA_PCM_LFLG_VALID) ) return -ENODEV;
  if ( !(pcm -> info_flags & ULTRA_PCM_INFO_MMAP) ) return -EIO;
  flags = ultra_pcm_file_flags( file -> f_flags );
  if ( ( vma -> vm_flags & ( VM_READ | VM_WRITE ) ) == ( VM_READ | VM_WRITE ) )
    return -EINVAL;
  if ( vma -> vm_flags & VM_READ )
    {
      if ( !(flags & ULTRA_PCM_LFLG_RECORD) ) return -EINVAL;
      pchn = &pcm -> record;
    }
   else
  if ( vma -> vm_flags & VM_WRITE )
    {
      if ( !(flags & ULTRA_PCM_LFLG_PLAY) ) return -EINVAL;
      pchn = &pcm -> playback;
    }
   else
    return -EINVAL;
  if ( vma -> vm_offset != 0 ) return -EINVAL;
  size = vma -> vm_end - vma -> vm_start;
  if ( size != pchn -> size )
    PRINTK( "ultra-pcm: wrong mmap() size 0x%x, should be 0x%x\n", (unsigned int)size, pchn -> size );
#if 0
  printk( "remap: to=0x%x from=0x%x size=0x%x\n", (int)vma -> vm_start, (int)virt_to_bus( pchn -> buffer ), (int)size );
#endif
  if ( remap_page_range( vma -> vm_start,
                         virt_to_bus( pchn -> buffer ), 
                         size,
                         vma -> vm_page_prot ) )
    return -EAGAIN;
#if LinuxVersionCode( 2, 1, 45 ) > LINUX_VERSION_CODE
  vma -> vm_inode = inode;
#if LinuxVersionCode( 2, 1, 43 ) > LINUX_VERSION_CODE
  inode -> i_count++;
#else
  atomic_inc( &inode -> i_count );
#endif
#else
  vma -> vm_dentry = dget( file -> f_dentry );
#endif
  pchn -> _dma -> mmaped = 1;
  pchn -> flags |= ULTRA_PCM_FLG_MMAP;
  ultra_pcm_fill_with_neutral( pchn );
  return 0;
}

/*
 *  Register functions
 */

ultra_pcm_t *ultra_pcm_new_device( ultra_card_t *card )
{
  ultra_pcm_t *pcm;

  if ( !card ) return NULL;
  pcm = (ultra_pcm_t *)ultra_malloc( sizeof( ultra_pcm_t ) );
  if ( !pcm ) return NULL;
  MEMSET( pcm, 0, sizeof( ultra_pcm_t ) );
  pcm -> flags = ULTRA_PCM_LFLG_NONE;
  pcm -> card = card;
  SLEEP_PREPARE( pcm, playback )
  SLEEP_PREPARE( pcm, record )
  return pcm;
}

int ultra_pcm_free( ultra_pcm_t *pcm )
{
  if ( !pcm ) return -EINVAL;
  if ( pcm -> playback.hw.private_free )
    pcm -> playback.hw.private_free( pcm -> playback.hw.private_data );
  if ( pcm -> record.hw.private_free )
    pcm -> record.hw.private_free( pcm -> record.hw.private_data );
  ultra_free( pcm, sizeof( ultra_pcm_t ) );
  return 0;  
}

int ultra_pcm_register( ultra_pcm_t *pcm, int pcm_device )
{
  int idx;

  if ( pcm_device < 0 || pcm_device > 1 ) return -EINVAL;
  idx = (pcm -> card -> number << 1) + pcm_device;
  if ( ultra_pcm_devices[ idx ] ) return -EBUSY;
  ultra_pcm_devices[ idx ] = pcm;
  if ( pcm -> flags & ULTRA_PCM_LFLG_USED )
    pcm -> flags |= ULTRA_PCM_LFLG_VALID;
  return 0;
}

int ultra_pcm_unregister( ultra_pcm_t *pcm )
{
  int idx;
  
  if ( !pcm ) return -EINVAL;
  for ( idx = 0; idx < ULTRA_CARDS * ULTRA_PCM_DEVICES; idx++ ) {
    if ( ultra_pcm_devices[ idx ] == pcm ) {
      ultra_pcm_devices[ idx ] = NULL;
    }
  }
  return ultra_pcm_free( pcm );
}

/*
 *  ENTRY functions
 */

static int ultra_pcm_unregister_callback( void );

static struct ultra_stru_register_minor ultra_pcm_reg = {
  "digital audio",
  
  1,					/* verify soundcard # */

  { ULTRA_MINOR_UPCM0, ULTRA_MINOR_UPCM1, -1 },
  { ULTRA_MINOR_PCM_8, ULTRA_MINOR_PCM_16, ULTRA_MINOR_AUDIO, ULTRA_MINOR_PCM1, -1 },
  
  ultra_pcm_unregister_callback,	/* unregister */
  
  NULL,					/* lseek */
  ultra_pcm_read,			/* read */
  ultra_pcm_write,			/* write */
  ultra_pcm_open,			/* open */
  ultra_pcm_release,			/* release */
#ifdef ULTRA_POLL
  ultra_pcm_poll,			/* poll */
#else
  ultra_pcm_select,			/* select */
#endif
  ultra_pcm_ioctl,			/* ioctl */
  ultra_pcm_mmap,			/* mmap */
};

static int ultra_pcm_unregister_callback( void )
{
  int err;
  
  if ( ( err = ultra_unregister_minor( &ultra_pcm_reg ) ) < 0 ) {
    PRINTD( "ultra-pcm: unregister error %d\n", err );
    return err;
  }
  return 0;
}
 
int init_module( void )
{
  int err;

  MEMSET( &ultra_pcm_devices, 0, sizeof( ultra_pcm_devices ) );
  if ( ( err = ultra_register_minor( &ultra_pcm_reg ) ) < 0 )
    return err;
  return 0;
}

void cleanup_module( void )
{
  ultra_pcm_unregister_callback();    
}
